#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-02-14
# @Filename: reconnect.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)


import asyncio

import click

from basecam.exceptions import CameraConnectionError

from ..tools import get_cameras
from .base import camera_parser


__all__ = ["reconnect"]


@camera_parser.command()
@click.argument("CAMERAS", nargs=-1, type=str, required=False)
@click.option(
    "--timeout",
    "-t",
    type=float,
    default=5.0,
    show_default=True,
    help="Seconds to wait until disconnect or reconnect command times out.",
)
async def reconnect(command, cameras, timeout):
    """Reconnects a camera."""

    cameras = get_cameras(command, cameras=cameras, fail_command=True)
    if not cameras:  # pragma: no cover
        return

    failed = False
    for camera in cameras:

        command.warning(text=f"reconnecting camera {camera.name!r}")

        try:
            await asyncio.wait_for(camera.disconnect(), timeout=timeout)
            command.info(
                camera=camera.name, text=f"camera {camera.name!r} was disconnected."
            )
        except CameraConnectionError as ee:
            command.warning(
                camera=camera.name,
                text=f"camera {camera.name!r} fail to disconnect: "
                f"{ee}. Will try to reconnect.",
            )
        except asyncio.TimeoutError:
            command.warning(
                camera=camera.name,
                text=f"camera {camera.name!r} timed out "
                "disconnecting. Will try to reconnect.",
            )

        try:
            await asyncio.wait_for(camera.connect(force=True), timeout=timeout)
            command.info(
                camera=camera.name, error=f"camera {camera.name!r} was reconnected."
            )
        except CameraConnectionError as ee:
            command.error(
                camera=camera.name,
                error=f"camera {camera.name!r} fail to reconnect: {ee}",
            )
            failed = True
        except asyncio.TimeoutError:
            command.error(
                camera=camera.name,
                error=f"camera {camera.name!r} timed out reconnecting.",
            )
            failed = True

    if failed:
        return command.fail("some cameras failed to reconnect.")
    else:
        return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-08-05
# @Filename: utils.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import asyncio
import logging
from contextlib import suppress

from sdsstools.logger import SDSSLogger


__all__ = ["LoggerMixIn", "Poller", "cancel_task"]


class LoggerMixIn(object):
    """A mixin to provide easy logging with a header."""

    log_header = ""
    logger: SDSSLogger

    def log(self, message, level=logging.DEBUG, use_header=True):
        """Logs a message with a header."""

        header = (self.log_header or "") if use_header else ""

        self.logger.log(level, header + message)


class Poller(object):
    """A task that runs a callback periodically.

    Parameters
    ----------
    name : str
        The name of the poller.
    callback : function or coroutine
        A function or coroutine to call periodically.
    delay : float
        Initial delay between calls to the callback.
    loop : event loop
        The event loop to which to attach the task.

    """

    def __init__(self, name, callback, delay=1, loop=None):

        self.name = name
        self.callback = callback

        self._orig_delay = delay
        self.delay = delay

        self.loop = loop or asyncio.get_event_loop()

        # Create two tasks, one for the sleep timer and another for the poller
        # itself. We do this because we want to be able to cancell the sleep
        # coroutine if we are going to change the delay.
        self._sleep_task = None
        self._task = None

    async def poller(self):
        """The polling loop."""

        while True:

            try:
                if asyncio.iscoroutinefunction(self.callback):
                    await self.callback()
                else:
                    self.callback()
            except Exception as ee:
                self.loop.call_exception_handler(
                    {"message": "failed running callback", "exception": ee}
                )
            self._sleep_task = self.loop.create_task(asyncio.sleep(self.delay))

            await self._sleep_task

    async def set_delay(self, delay=None, immediate=False):
        """Sets the delay for polling.

        Parameters
        ----------
        delay : float
            The delay between calls to the callback. If `None`, restores the
            original delay.
        immediate : bool
            If `True`, stops the currently running task and sets the
            new delay. Otherwise waits for the current task to complete.

        """

        # Only change delay if the difference is significant.
        if delay and abs(self.delay - delay) < 1e-6:
            return

        if not self.running:
            return

        if immediate:
            await self.stop()
            self.start(delay)
        else:
            self.delay = delay or self._orig_delay

    def start(self, delay=None):
        """Starts the poller.

        Parameters
        ----------
        delay : float
            The delay between calls to the callback. If not specified,
            restores the original delay used when the class was instantiated.

        """

        self.delay = delay or self._orig_delay

        if self.running:
            return

        self._task = self.loop.create_task(self.poller())

        return self

    async def stop(self):
        """Cancel the poller."""

        if not self.running:
            return

        if self._task:
            self._task.cancel()
            with suppress(asyncio.CancelledError):
                await self._task

    async def call_now(self):
        """Calls the callback immediately."""

        restart = False
        delay = self.delay
        if self.running:
            await self.stop()
            restart = True

        if asyncio.iscoroutinefunction(self.callback):
            await self.loop.create_task(self.callback())
        else:
            self.callback()

        if restart:
            self.start(delay=delay)

    @property
    def running(self):
        """Returns `True` if the poller is running."""

        if self._task and not self._task.cancelled():
            return True

        return False


async def cancel_task(task):
    """Cleanly cancels a task."""

    if task and not task.done():
        task.cancel()
        with suppress(asyncio.CancelledError):
            await task



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-08-06
# @Filename: actor.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import logging
import os

from typing import List, Optional, Union

import clu.parsers.click
from clu import BaseActor, JSONActor
from clu.tools import ActorHandler
from sdsstools.logger import SDSSLogger

from basecam import CameraSystem, EventListener
from basecam.exceptions import CameraWarning

from . import commands


__all__ = ["BaseCameraActor", "CameraActor"]


class BaseCameraActor(BaseActor):
    """Base class for a camera CLU-like actor class.

    Expands a `CLU <https://clu.readthedocs.io/en/latest/>`__ actor
    to receive commands, interact with the camera system, and reply to
    the commander. This base class needs to be subclassed along with the
    desired implementation of CLU `~clu.actor.BaseActor`. For example ::

        from clu.actor import AMQPActor

        class MyCameraActor(BaseCameraActor, AMQPActor):
            pass

    Parameters
    ----------
    camera_system
        The camera system, already instantiated.
    default_cameras
        A list of camera names or UIDs that define what cameras to use by
        default in most command.
    command_parser
        The list of commands to use. It must be a command group deriving from
        `~clu.parsers.click.CluGroup` containing all the commands to use. If
        ``commands=None``, uses the internal command set.
    schema
        The path to the JSONSchema file with the actor datamodel. If ``"internal"``,
        uses the default ``basecam`` model; `None` disables model validation.
    args,kwars
        Arguments and keyword arguments to be passed to the actor class.
    """

    def __init__(
        self,
        camera_system: CameraSystem,
        *args,
        default_cameras: Union[List[str], str, None] = None,
        command_parser: clu.parsers.click.CluGroup = None,
        schema: Optional[str] = "internal",
        **kwargs,
    ):

        self._check_is_subclass()

        assert camera_system is not None
        self.camera_system = camera_system

        # Add actor to camera class context and all connected cameras.
        self.camera_system.camera_class.fits_model.context.update({"__actor__": self})
        for camera in self.camera_system.cameras:
            camera.fits_model.context.update({"__actor__": self})

        #: An `.EventListener` that can be used to wait or respond to events.
        self.listener = EventListener()
        self.camera_system.notifier.register_listener(self.listener)

        self.log: SDSSLogger
        self.parser = command_parser or commands.camera_parser

        if schema == "internal":
            schema = os.path.join(os.path.dirname(__file__), "schema.json")

        super().__init__(*args, schema=schema, **kwargs)

        # Add commands that depend on what mixins the base camera has
        # been subclassed with.
        if self.parser == commands.camera_parser:
            self._add_optional_commands()

        # Output camera_system log messages as keywords.
        actor_handler = ActorHandler(
            self,
            code_mapping={logging.INFO: "d"},
            filter_warnings=[CameraWarning, UserWarning],
        )
        actor_handler.setLevel(logging.INFO)
        self.camera_system.logger.addHandler(actor_handler)

        self.default_cameras = None
        self.set_default_cameras(default_cameras)

    def _check_is_subclass(self):
        """Checks if the object is a subclass of a CLU actor."""

        error = "BaseCameraActor must be sub-classed along with a CLU actor class."
        bases = self.__class__.__bases__

        assert issubclass(self.__class__, BaseCameraActor), error

        # Check that at least one of the bases is a sublass of BaseActor
        for base in bases:
            if base == BaseCameraActor:
                continue
            if issubclass(base, BaseActor):
                return

        raise RuntimeError(error)

    def _add_optional_commands(self):
        """Adds commands and groups based on the mixins present."""

        camera_class = self.camera_system.camera_class

        for mixin in camera_class.__bases__:
            mixin_name = mixin.__name__
            if mixin_name in commands._MIXIN_TO_COMMANDS:
                for command in commands._MIXIN_TO_COMMANDS[mixin_name]:
                    self.parser.add_command(command)

    def set_default_cameras(self, cameras: Union[str, List[str], None] = None):
        """Sets the camera(s) that will be used by default.

        These cameras will be used by default when a command is issued without
        listing the cameras to command.

        Parameters
        ----------
        cameras : str or list
            A list of camera names or a string with comma-separated camera
            names. If `None`, no cameras will be considered default and
            commands will fail if they do not specify a camera.

        """

        if cameras is None:
            self.default_cameras = None
            return

        if isinstance(cameras, str):
            self.default_cameras = cameras.split(",")
        elif isinstance(cameras, (list, tuple)):
            self.default_cameras = list(cameras)
        else:
            raise ValueError(f"invalid data type for cameras={cameras!r}")

        connected_cameras = [camera.name for camera in self.camera_system.cameras]
        for camera in self.default_cameras:
            if camera not in connected_cameras:
                self.log.warning(
                    f"camera {camera!r} made default but is not connected."
                )


class CameraActor(BaseCameraActor, JSONActor):
    """A camera actor that replies with JSONs using `~clu.actor.JSONActor`."""

    pass



# encoding: utf-8
# flake8: noqa

from sdsstools import get_package_version

from .camera import *
from .events import *
from .exceptions import *
from .exposure import *
from .notifier import *


NAME = "sdss-basecam"

__version__ = get_package_version(__file__, "sdss-basecam") or "dev"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-02-14
# @Filename: base.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import click

from clu.parsers.click import CluGroup, help_, ping


__all__ = ["camera_parser", "list_"]


@click.group(cls=CluGroup)
def camera_parser():
    pass


camera_parser.add_command(help_)
camera_parser.add_command(ping)


@camera_parser.command(name="list")
async def list_(command):
    """Lists cameras connected to the camera system."""

    cameras = [camera.name for camera in command.actor.camera_system.cameras]
    command.finish(cameras=cameras, default_cameras=command.actor.default_cameras)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-02-16
# @Filename: __doc_parser.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

# This parser is only for documentation purposes and should not be used.

import copy

from . import area, binning, camera_parser, shutter, temperature  # type: ignore


__doc_parser = copy.deepcopy(camera_parser)
__doc_parser.add_command(shutter)
__doc_parser.add_command(temperature)
__doc_parser.add_command(binning)
__doc_parser.add_command(area)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-01-10
# @Filename: exposure.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import functools
import os
import pathlib
import re

from typing import Callable, List, Optional, Tuple, Union

import astropy
import astropy.time
import astropy.wcs
import numpy
from astropy.io.fits import BinTableHDU, HDUList, ImageHDU

import basecam.camera
import basecam.models

from .exceptions import ExposureError


__all__ = ["Exposure", "ImageNamer"]


class Exposure(object):
    """Exposure class. Represents an exposure data and its metadata.

    An `.Exposure` is defined by the raw image taken by the camera, a series
    of attributes that define the exposure (exposure time, shutter, etc.) and
    a data model that is used to generate the FITS file for the exposure.

    Parameters
    ----------
    camera
        The instance of a subclass of `.BaseCamera` that took this exposure.
    filename
        The filename of the FITS image generated.
    data
        The exposure raw data, as a 2D array.
    fits_model
        The `model <.FITSModel>` to create the FITS image. If `None`, a single
        extension with a basic header will be used.
    wcs
        The WCS object describing the astrometry of this exposure.

    Attributes
    ----------
    image_type
        The type of image, one of ``bias``, ``flat``, ``dark``, ``object``.
    exptime : float
        The exposure time, in seconds, of a single integration.
    exptime_n : float
        The total exposure time, in seconds. If the image is stacked, this is
        the total time, i.e., the sum of the exposure time of each stacked
        image.
    stack : int
        Number of exposures stacked.
    stack_function
        Name of the function used for stacking.
    filename
        The path where to write the image.
    wcs : ~astropy.wcs.WCS
        The WCS object describing the astrometry of this exposure.
    """

    def __init__(
        self,
        camera: basecam.camera.BaseCamera,
        filename: Optional[str] = None,
        data: Optional[numpy.ndarray] = None,
        fits_model: Optional[basecam.models.FITSModel] = None,
        wcs: Optional[astropy.wcs.WCS] = None,
    ):

        self.camera = camera
        self.data = data
        self.fits_model = fits_model.copy() if fits_model else None
        self.filename = filename

        self._obstime: astropy.time.Time = astropy.time.Time.now()

        self.exptime: Optional[float] = None
        self.exptime_n: Optional[float] = None
        self.stack: int = 1
        self.stack_function: Optional[Callable[..., numpy.ndarray]] = None
        self.image_type: Optional[str] = None

        self._extra_hdus: List[Tuple[Union[BinTableHDU, ImageHDU], Optional[int]]] = []

        self.wcs = wcs

    @property
    def obstime(self) -> astropy.time.Time:
        """The time at the beginning of the observation.

        It must be an `astropy.time.Time` object or a datetime in ISO
        format; in the latter case, UTC scale is assumed.
        """

        return self._obstime

    @obstime.setter
    def obstime(self, value: astropy.time.Time):

        if isinstance(value, astropy.time.Time):
            self._obstime = value
        elif isinstance(value, str):
            self._obstime = astropy.time.Time(value, format="iso", scale="utc")
        else:
            raise ExposureError(f"invalid obstime {value}")

    def add_hdu(self, hdu: Union[BinTableHDU, ImageHDU], index: Optional[int] = None):
        """Adds an HDU to the list of extensions.

        Parameters
        ----------
        hdu
            The `~astropy.io.fits.BinTableHDU` or `~astropy.io.fits.ImageHDU` HDU to
            append.
        index
            The index where the extension will be added. Extra HDUs are inserted in
            order after the FITS model has been generated. ``index=None`` appends the
            new HDU at the end of the list. Note that ``astropy.io.fits`` may change
            the final order of the extensions to ensure that a primary HDU remains
            as the first HDU.
        """

        assert isinstance(hdu, (BinTableHDU, ImageHDU)), "invalid HDU type"

        self._extra_hdus.append((hdu, index))

    def to_hdu(self, context={}) -> HDUList:
        """Return an `~astropy.io.fits.HDUList` for the image.

        Parameters
        ----------
        context
            A dictionary of arguments used to evaluate the FITS model.

        Returns
        -------
        hdulist
            A list of HDUs in which the FITS data model has been evaluated
            for this exposure.
        """

        fits_model = self.fits_model or basecam.models.FITSModel()

        hdulist = fits_model.to_hdu(self, context=context)
        for hdu, index in self._extra_hdus:
            if index:
                hdulist.insert(index, hdu)
            else:
                hdulist.append(hdu)

        return hdulist

    async def write(
        self,
        filename: Optional[str] = None,
        context={},
        overwrite=False,
        checksum=True,
    ) -> HDUList:
        """Writes the image to disk.

        Parameters
        ----------
        filename
            The path where to write the file. If not provided, uses
            ``Exposure.filename``.
        context
            A dictionary of arguments used to evaluate the FITS model.
        overwrite
            Whether to overwrite the image if it exists.
        checksum
            When `True` adds both ``DATASUM`` and ``CHECKSUM`` cards to the
            headers of all HDUs written to the file.

        Returns
        -------
        hdulist
            A list of HDUs in which the FITS data model has been evaluated
            for this exposure.
        """

        filename = filename or self.filename
        if not filename:
            raise ExposureError("filename not set.")

        hdulist = self.to_hdu(context=context)

        dirname = os.path.realpath(os.path.dirname(filename))
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        writeto_partial = functools.partial(
            hdulist.writeto, filename, overwrite=overwrite, checksum=checksum
        )

        loop = asyncio.get_event_loop()
        await loop.run_in_executor(None, writeto_partial)

        return hdulist


class ImageNamer(object):
    """Creates a new sequential filename for an image.

    Parameters
    ----------
    basename
        The basename of the image filenames. Must contain a placeholder
        ``num`` in the place where to insert the sequence number. For example,
        ``'test-{num:04d}.fits'`` will produce image names ``test-0001.fits``,
        ``test-0002.fits``, etc. It's also possible to use placeholders for
        camera values, e.g. ``{camera.name}-{num}.fits``.
    dirname
        The directory for the images. Can include an expression based on the
        ``date`` substitution which is a `~astropy.time.Time.now` object. For
        example: ``dirname='/data/{camera.uid}/{int(date.mjd)}'``.
    overwrite
        If `True`, the sequence will start at 1 regardless of the existing
        images. If `False`, the first element in the sequence will be selected
        to avoid colliding with any image already existing in the directory.
    camera
        A `.BaseCamera` instance. It can also be passed when calling the
        instance.

    Examples
    --------
    >>> namer = ImageNamer('{camera.name}-{num:04d}.fits', dirname='testdir')
    >>> namer(camera=camera)
    PosixPath('testdir/my_camera-0001.fits')
    >>> namer(camera=camera)
    PosixPath('testdir/my_camera-0002.fits')
    """

    def __init__(
        self,
        basename: str = "{camera.name}-{num:04d}.fits",
        dirname: str = ".",
        overwrite: bool = False,
        camera: Optional[basecam.camera.BaseCamera] = None,
    ):

        assert re.match(r".+(\{num.+\}).+", basename), "invalid basename."

        self._basename: str
        self.basename = basename

        self.dirname: Union[pathlib.Path, str] = pathlib.Path(dirname)
        self.overwrite: bool = overwrite

        self._last_num = 0
        self.camera = camera

    @property
    def basename(self) -> str:
        """The image name pattern."""
        return self._basename

    @basename.setter
    def basename(self, value: str):
        """Sets the basename."""
        # We want to expand everything except the num first so we "double-escape" it.
        self._basename = re.sub(r"(\{num.+\})", r"{\1}", value)

    def get_dirname(self) -> pathlib.Path:
        """Returns the evaluated dirname."""

        date = astropy.time.Time.now()

        return pathlib.Path(
            eval(f'f"{self.dirname}"', {}, {"date": date, "camera": self.camera})
        )

    def _get_num(self, basename: str) -> int:
        """Returns the counter value."""

        if self.overwrite:
            return self._last_num + 1

        regex = re.compile(re.sub(r"\{num.+\}", "(?P<num>[0-9]*?)", basename))

        dirname = self.get_dirname()
        all_files = list(map(str, dirname.glob("*")))

        match_files = list(filter(regex.search, all_files))

        if len(match_files) == 0:
            return self._last_num + 1

        values = [int(regex.search(file).group(1)) for file in match_files]
        return max(values) + 1

    def __call__(
        self,
        camera: Optional[basecam.camera.BaseCamera] = None,
        update_num: bool = True,
    ) -> pathlib.Path:

        camera = camera or self.camera

        if camera:
            expanded_basename = self.basename.format(camera=camera)
        else:
            expanded_basename = self.basename.format()

        num = self._get_num(expanded_basename)
        path = self.get_dirname() / expanded_basename.format(num=num)

        if update_num:
            self._last_num = num

        return path



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-08-06
# @Filename: commands.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import json
import os

from typing import Any, Dict


__all__ = ["get_cameras", "get_schema"]


def get_cameras(command, cameras=None, check_cameras=True, fail_command=False):
    """A helper to determine what cameras to use.

    Parameters
    ----------
    command
        The command that wants to access the cameras.
    cameras : list
        The list of camera names passed to the command.
    check_cameras : bool
        If any of the cameras is not connected, returns `False`.
    fail_command : bool
        Fails the command before returning `False`.

    Returns
    -------
    cameras : list of `.Camera` instances
        A list of `.Camera` instances that match the input ``cameras`` or,
        if ``cameras=None``, the default cameras. If ``cameras=None`` and
        there are no default cameras defined, returns all the connected
        camera. If ``check_cameras=True`` and any of the selected cameras
        is not connected, the command is failed and returns `False`.

    """

    default = command.actor.default_cameras

    if not cameras:
        if default is None or len(default) == 0:
            camera_instances = command.actor.camera_system.cameras
            cameras = [camera_instance.name for camera_instance in camera_instances]
        else:
            cameras = default

    camera_instances = []
    for camera in cameras:
        camera_instance = command.actor.camera_system.get_camera(name=camera)
        if camera_instance is False:
            if fail_command:
                command.fail(text=f"camera {camera} is not connected.")
            return False
        camera_instances.append(camera_instance)

    if check_cameras:
        for camera_instance in camera_instances:
            if not camera_instance.connected:
                if fail_command:
                    command.fail(
                        text=f"camera {camera_instance.name} "
                        "has not been initialised."
                    )
                return False

    if len(cameras) == 0:
        if fail_command:
            command.fail(text="no cameras connected.")
        return False

    return camera_instances


def get_schema() -> Dict[str, Any]:
    """Returns the actor schema as a dictionary."""

    schema = json.loads(
        open(os.path.join(os.path.dirname(__file__), "schema.json")).read()
    )

    return schema



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-02-10
# @Filename: card.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import abc
import re
import warnings
from collections.abc import Iterable
from contextlib import contextmanager

from typing import Any, Dict, List, NamedTuple, Optional, Union, cast

import astropy.io.fits
import astropy.wcs

from ..exceptions import CardError, CardWarning
from ..exposure import Exposure


__all__ = [
    "Card",
    "DefaultCard",
    "CardGroup",
    "MacroCard",
    "WCSCards",
    "DEFAULT_CARDS",
]


class EvaluatedCard(NamedTuple):
    """A tuple describing the result of evaluating a `.Card`."""

    name: str
    value: Any
    comment: str


class Card(object):
    """A card representing a single entry in the header.

    This class is similar to astropy's `~astropy.io.fits.Card` with the
    difference that the value of the card does not need to be a literal and is
    evaluated when `.evaluate` is called.

    The value of the card can be:

    - A Python literal.

    - A callable, with a list of arguments, that is called when the card
      is evaluated.

    - A string using Python's :ref:`python:formatstrings`. As with normal
      strings, the replacement fields are surrounded by ``{`` and ``}`` and
      filled when `.evaluate` is called. The replacement fields can point to an
      attribute, function, method, or property. In addition, it is possible to
      use the string ``__exposure__`` and ``__camera__`` to refer to the
      `.Exposure` instance for which this card will be evaluated, and the
      instance of `.BaseCamera` that took the image, respectively. For example,
      ``value='{__camera__.name}'`` will be replaced with the name of the
      camera, and ``value='{__exposure__.obstime}'`` will retrieve the time of
      the start of the exposure.

    - A string to be evaluated using Python's `eval` function. For example,
      ``value='__camera__.get_temperature()'``.

    If the ``name`` of the card is one of the :ref:`Default Cards <default-cards>`,
    the value and comment are automatically defined and do not need to be
    specified.

    Parameters
    ----------
    name
        The name of the card. Will be trimmed to 8 characters. Can be one of
        the :ref:`Default Cards <default-cards>` names, which defines the value
        and comment.
    value
        A Python literal, callable, or string, as described above.
    comment
        A short comment describing the card value.
    default
        Default value to use if the card value contains placeholders that are not
        provided while the card is evaluated.
    type
        The type of the card value. The value will be casted after processing.
    autocast
        If `True` and ``type`` is not defined, tries to cast the value.
    fargs
        If ``value`` is a callable, the arguments to pass to it when it gets
        called. The arguments are evaluated following the same rules as with
        the ``value`` before being passed to the callable.
    evaluate
        If `True`, assumes the value is a string to be evaluated in realtime.
        The context is used as ``locals`` for the evaluation. For security,
        ``globals`` are not available.
    context
        A dictionary of parameters used to fill the value replacement fields.
        Two values, ``__exposure__`` and ``__camera__``, are always defined.
        This context can be updated during the evaluation of the card.
    """

    def __new__(cls, name: Union[str, Iterable[Any]], *args, **kwargs):

        if isinstance(name, str):
            if cls == Card and name.upper() in DEFAULT_CARDS:
                if len(args) == 0 and len(kwargs) == 0:
                    return DEFAULT_CARDS[name.upper()]
            return super().__new__(cls)

        return cls(*name)

    def __init__(
        self,
        name: Union[str, Iterable[Any]],
        value: Any = "",
        comment: str = "",
        type: Optional[type] = None,
        autocast: bool = True,
        default: Any = None,
        fargs: Optional[tuple] = None,
        evaluate: bool = False,
        context: Dict[str, Any] = {},
    ):

        if hasattr(self, "name"):
            return

        assert isinstance(name, str)

        self.name = name
        if len(self.name) > 8:
            warnings.warn(f"trimming {self.name} to 8 characters", CardWarning)
            self.name = self.name[:8]

        self.value = value
        self.comment = comment

        self._default = default
        self._type = type
        self._autocast = autocast
        self._fargs = fargs
        self._evaluate = evaluate

        self.context = context or {}

    def __repr__(self):
        return f"<{self.__class__.__name__} (name={self.name!r}, value={self.value!r})>"

    @contextmanager
    def set_exposure(self, exposure, context={}):
        """Sets the current exposure and context, clearing it on exit."""

        original_context = self.context.copy()

        try:
            self._exposure = exposure
            self.context.update({"__exposure__": self._exposure})
            if hasattr(self._exposure, "camera") and self._exposure.camera:
                self.context.update({"__camera__": self._exposure.camera})
            self.context.update(context)
            yield
        finally:
            self._exposure = None
            self.context = original_context

    def _render_value(self, value):
        """Returns the value after evaluating the string template."""

        if not isinstance(value, str):
            return value

        if self._evaluate:
            return eval(value, {}, self.context)

        # Get a list of the placeholders, if any.
        placeholders = re.findall(r"{([0-9a-zA-Z_]+)\.?.*?}", value)

        if not all([p in self.context for p in placeholders]):
            if self._default:
                return self._default
            else:
                raise ValueError(
                    "The context does not include all the Card value placeholders."
                )

        # Expand the placeholders. If there are none, no harm done.
        return value.format(**self.context)

    def _evaluate_callable(self, func, use_fargs=True):
        """Evaluates a value that is a callable function or method."""

        if use_fargs and self._fargs:
            # Evaluate the fargs as values so that we can apply
            # all the magic to them as well.
            fargs = (
                self._evaluate_value(farg, use_fargs=False) for farg in self._fargs
            )
            return func(*fargs)

        return func()

    def _evaluate_value(self, value, use_fargs=True):
        """Evaluates a value, expanding its template."""

        if callable(value):
            return self._evaluate_callable(value, use_fargs=use_fargs)

        return self._render_value(value)

    def evaluate(
        self,
        exposure: Exposure,
        context: Dict[str, Any] = {},
    ) -> EvaluatedCard:
        """Evaluates the card.

        Evaluates the card value, expanding its template parameters. If the
        value is a callable, calls the function at this time.

        Parameters
        ----------
        exposure
            The exposure for which we want to evaluate the card.
        context
            A dictionary of arguments used to evaluate the parameters in
            the value. These argument update those defined when the `.Card`
            was created.

        Returns
        -------
        EvaluatedCard
            A tuple with the name, evaluated value, and comment for this card.
        """

        with self.set_exposure(exposure, context):
            try:
                rendered_value = self._evaluate_value(self.value)
            except BaseException:
                if self._default:
                    rendered_value = self._default
                else:
                    raise

        if self._type:
            rendered_value = self._type(rendered_value)
        elif self._autocast and isinstance(rendered_value, str):
            rendered_value = cast(str, rendered_value)
            if re.match("^true$", rendered_value, re.IGNORECASE):
                rendered_value = True
            elif re.match("^false$", rendered_value, re.IGNORECASE):
                rendered_value = False
            elif re.match("^none$", rendered_value, re.IGNORECASE):
                rendered_value = self._default if self._default else "None"
            else:
                try:
                    rendered_value = int(rendered_value)
                except ValueError:
                    try:
                        rendered_value = float(rendered_value)
                    except ValueError:
                        pass

        return EvaluatedCard(self.name, rendered_value, self.comment)


class CardGroup(list):
    """A group of `.Card` instances.

    A `.CardGroup` is just a list of `Cards <.Card>` that are grouped for
    convenience and for easy reuse.

    Parameters
    ----------
    cards
        A list of `.Card` instances. Elements can also be a tuple of two or
        three elements (for name, value, and optionally a comment) or a string with
        a default card.
    name
        A name for the card group.
    use_group_title
        Whether to prepend a COMMENT card with the group title when creating
        a header.
    """

    name: Optional[str] = None

    def __init__(
        self,
        cards: Iterable[Union[Card, Iterable, str]],
        name: Optional[str] = None,
        use_group_title: bool = True,
    ):

        self.name = name or self.name
        self.use_group_title: bool = use_group_title

        cards = [self._process_input(card) for card in cards or []]
        list.__init__(self, cards)

    def __repr__(self):

        return f"<CardGroup {list.__repr__(self)!s}>"

    def _process_input(self, card: Union[Card, Iterable]):
        """Processes the input and converts it into a valid card."""

        if isinstance(card, Card):
            return card
        elif isinstance(card, (Iterable, str)):
            return Card(card)
        else:
            raise CardError(f"invalid card {card}")

    def append(self, card: Card):
        list.append(self, self._process_input(card))

    def insert(self, idx, card: Card):
        list.insert(self, idx, self._process_input(card))

    def evaluate(self, exposure, context={}):
        """Evaluates all the cards.

        Parameters
        ----------
        exposure : .Exposure
            The exposure for which we want to evaluate the cards.
        context : dict
            A dictionary of arguments used to evaluate the cards.

        Returns
        -------
        list
            A list of tuples with the name, evaluated value, and comment
            for each card.
        """

        return [card.evaluate(exposure, context=context) for card in self]

    def to_header(
        self,
        exposure: Exposure,
        context: Dict[str, Any] = {},
        use_group_title: bool = False,
    ) -> astropy.io.fits.Header:
        """Evaluates all the cards and returns a header.

        Parameters
        ----------
        exposure
            The exposure for which we want to evaluate the cards.
        context
            A dictionary of arguments used to evaluate the cards.
        use_group_title
            Whether to prepend a COMMENT card with the group title when
            creating the header.

        Returns
        -------
        header
            A header composed from the cards in the group.
        """

        header = astropy.io.fits.Header(self.evaluate(exposure, context=context))

        use_group_title = use_group_title or self.use_group_title
        if use_group_title and self.name:
            header.insert(0, ("COMMENT", "{s:#^30}".format(s=f" {self.name} ")))

        return header


class MacroCard(object, metaclass=abc.ABCMeta):
    """A macro that returns a list of keywords and values.

    This is an abstract class in which the `.macro` method must be overridden
    to return a list of keywords and values that can be used to create a
    header.

    Parameters
    ----------
    name
        A name for the macro group.
    use_group_title
        Whether to prepend a COMMENT card with the macro title when creating
        a header.
    kwargs
        Additional arguments for the macro.
    """

    name = None

    def __init__(
        self,
        name: Optional[str] = None,
        use_group_title: bool = False,
        **kwargs,
    ):

        self.name = name or self.name
        self.use_group_title = use_group_title
        self.kwargs = kwargs

    def __repr__(self):

        return f"<{self.__class__.__name__} (name={self.name})>"

    @abc.abstractmethod
    def macro(self, exposure: Exposure, context: Dict[str, Any] = {}):
        """The macro.

        Must return a list of item which can be tuples with the format
        ``(keyword, value, comment)`` or ``(keyword, value)``, `.Card` instances,
        or `.CardGroup` instances. Cards and card groups will be evaluated.

        Parameters
        ----------
        exposure
            The exposure for which we want to evaluate the macro.
        context
            A dictionary of parameters that can be used by the macro.
        """

        raise NotImplementedError

    def evaluate(self, exposure: Exposure, context: Dict[str, Any] = {}) -> List[tuple]:
        """Evaluates the macro. Equivalent to calling `.macro` directly.

        Parameters
        ----------
        exposure
            The exposure for which we want to evaluate the macro.
        context
            A dictionary of parameters that can be used by the macro.

        Returns
        -------
        tuples
            A list of tuples with the format ``(keyword, value, comment)``
            or ``(keyword, value)``.
        """

        cards = self.macro(exposure, context=context)
        new_cards = []
        for card in cards:
            if isinstance(card, Card):
                new_cards.append(card.evaluate(exposure, context=context))
            elif isinstance(card, CardGroup):
                new_cards += card.evaluate(exposure, context=context)
            else:
                new_cards.append(card)
        return new_cards

    def to_header(
        self,
        exposure: Exposure,
        context: Dict[str, Any] = {},
        use_group_title=False,
    ) -> astropy.io.fits.Header:
        """Evaluates the macro and returns a header.

        Parameters
        ----------
        exposure
            The exposure for which we want to evaluate the macro.
        context
            A dictionary of arguments used to evaluate the macro.
        use_group_title
            Whether to prepend a COMMENT card with the macro title when
            creating the header.

        Returns
        -------
        header
            A header composed from the cards in the macro.
        """

        header = astropy.io.fits.Header(self.evaluate(exposure, context=context))

        use_group_title = use_group_title or self.use_group_title
        if use_group_title and self.name:
            header.insert(0, ("COMMENT", "{s:#^30}".format(s=f" {self.name} ")))

        return header


class WCSCards(MacroCard):
    """A macro that adds WCS header information.

    To use, just add ``WCSCards()`` to the header model and make sure the
    ``Exposure.wcs`` is set. If ``Exposure.wcs=None``, a default WCS header will
    be added.
    """

    name = "WCS information"

    def macro(self, exposure: Exposure, context: Dict[str, Any] = {}):
        if exposure.wcs is None:
            wcs = astropy.wcs.WCS()
        else:
            wcs = exposure.wcs
        return list(wcs.to_header().cards)


class DefaultCard(Card):
    pass


#: Default cards
DEFAULT_CARDS: Dict[str, DefaultCard] = {
    "EXPTIME": DefaultCard(
        "EXPTIME",
        value="{__exposure__.exptime}",
        comment="Exposure time of single integration [s]",
        type=float,
    ),
    "EXPTIMEN": DefaultCard(
        "EXPTIMEN",
        value="{__exposure__.exptime_n}",
        comment="Total exposure time [s]",
        default=-999,
    ),
    "STACK": DefaultCard(
        "STACK",
        value="{__exposure__.stack}",
        comment="Number of stacked frames",
        default=1,
        type=int,
    ),
    "STACKFUN": DefaultCard(
        "STACKFUN",
        value="{__exposure__.stack_function.__name__}",
        comment="Function used for stacking",
        default="NA",
    ),
    "OBSTIME": DefaultCard(
        "OBSTIME",
        value="{__exposure__.obstime.tai}",
        comment="Time of the start of the exposure [TAI]",
    ),
    "IMAGETYP": DefaultCard(
        "IMAGETYP",
        value="{__exposure__.image_type}",
        comment="The image type of the file",
    ),
    "CAMNAME": DefaultCard(
        "CAMNAME",
        value="{__camera__.name}",
        comment="Camera name",
        default="NA",
    ),
    "CAMUID": DefaultCard(
        "CAMUID",
        value="{__camera__.uid}",
        comment="Camera UID",
        default="NA",
    ),
    "VCAM": DefaultCard(
        "VCAM",
        value="{__camera__.__version__}",
        comment="Version of the camera library",
        default="NA",
    ),
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-01-12
# @Filename: builtin.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from .card import Card
from .fits import Extension, FITSModel, HeaderModel


__all__ = ["basic_header_model", "basic_fits_model", "basic_fz_fits_model"]

#: A basic header model with camera and exposure information.
basic_header_model = HeaderModel(
    [
        "VCAM",
        "CAMNAME",
        "CAMUID",
        "IMAGETYP",
        "EXPTIME",
        "EXPTIMEN",
        "STACK",
        "STACKFUN",
        Card(
            "TIMESYS",
            value="TAI",
            comment="The time scale system",
        ),
        Card(
            "DATE-OBS",
            value="{__exposure__.obstime.tai.isot}",
            comment="Date (in TIMESYS) the exposure started",
        ),
    ]
)


#: A basic FITS model for uncompressed images. Includes a single extension
#: with the raw data and a `.basic_header_model`.
basic_fits_model = FITSModel(
    [Extension(data="raw", header_model=basic_header_model, name="PRIMARY")]
)


#: A compressed, basic FITS model. Similar to `.basic_fits_model` but uses
#: ``RICE_1`` compression.
basic_fz_fits_model = FITSModel(
    [
        Extension(
            data="raw",
            header_model=basic_header_model,
            compressed="RICE_1",
            name="PRIMARY",
        )
    ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-02-14
# @Filename: temperature.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import asyncio

import click

from clu.parsers.click import CluCommand

from basecam.exceptions import CameraError

from ..tools import get_cameras


__all__ = ["temperature"]


@click.command(cls=CluCommand)
@click.argument("CAMERAS", nargs=-1, type=str, required=False)
@click.argument("TEMPERATURE", type=float, required=False)
async def temperature(command, cameras, temperature):
    """Controls the camera temperature.

    If called without a temperature value, returns the current temperature.
    """

    cameras = get_cameras(command, cameras=cameras, fail_command=True)
    if not cameras:  # pragma: no cover
        return

    temperature_tasks = []

    for camera in cameras:
        if not temperature:
            command.info(
                temperature=dict(
                    camera=camera.name,
                    ccd_temp=await camera.get_temperature(),
                )
            )
        else:
            temperature_tasks.append(camera.set_temperature(temperature))

    if not temperature:
        return command.finish()

    results = await asyncio.gather(*temperature_tasks, return_exceptions=True)
    failed = False
    for ii, result in enumerate(results):
        if isinstance(result, CameraError):
            command.error(error=dict(camera=cameras[ii].name, error=str(result)))
            failed = True
        else:
            command.info(
                temperature=dict(
                    camera=cameras[ii].name,
                    ccd_temp=await cameras[ii].get_temperature(),
                )
            )

    if failed:
        return command.fail("one or more cameras failed to set temperature.")
    else:
        return command.finish("all cameras have reached their set points.")
